@@ -57,9 +57,6 @@
 #endif
 
 #include "sasl_defs.h"
-#ifdef TLS
-#include <openssl/ssl.h>
-#endif
 
 /* for NAPI pinning feature */
 #ifndef SO_INCOMING_NAPI_ID
@@ -362,8 +359,7 @@ struct slab_stats {
     X(proxy_conn_requests) \
     X(proxy_conn_errors) \
     X(proxy_conn_oom) \
-    X(proxy_req_active) \
-    X(proxy_await_active)
+    X(proxy_req_active)
 #endif
 
 /**
@@ -418,6 +414,7 @@ struct stats {
     uint64_t      extstore_compact_resc_old; /* items re-written during compaction */
 #endif
 #ifdef TLS
+    uint64_t      ssl_proto_errors; /* TLS failures during SSL_read() and SSL_write() calls */
     uint64_t      ssl_handshake_errors; /* TLS failures at accept/handshake time */
     uint64_t      ssl_new_sessions; /* successfully negotiated new (non-reused) TLS sessions */
 #endif
@@ -484,6 +481,7 @@ struct settings {
     bool lru_maintainer_thread; /* LRU maintainer background thread */
     bool lru_segmented;     /* Use split or flat LRU's */
     bool slab_reassign;     /* Whether or not slab reassignment is allowed */
+    bool ssl_enabled; /* indicates whether SSL is enabled */
     int slab_automove;     /* Whether or not to automatically move slabs */
     double slab_automove_ratio; /* youngest must be within pct of oldest */
     unsigned int slab_automove_window; /* window mover for algorithm */
@@ -527,8 +525,7 @@ struct settings {
     unsigned int ext_global_pool_min;
 #endif
 #ifdef TLS
-    bool ssl_enabled; /* indicates whether SSL is enabled */
-    SSL_CTX *ssl_ctx; /* holds the SSL server context which has the server certificate */
+    void *ssl_ctx; /* holds the SSL server context which has the server certificate */
     char *ssl_chain_cert; /* path to the server SSL chain certificate */
     char *ssl_key; /* path to the server key */
     int ssl_verify_mode; /* client certificate verify mode */
@@ -755,7 +752,7 @@ typedef struct {
     int proxy_vm_last_kb;
     unsigned int proxy_vm_negative_delta;
     int proxy_vm_gcrunning;
-    bool proxy_vm_needspoke;
+    int proxy_vm_gcpokemem;
     uint64_t proxy_active_req_limit;
     uint64_t proxy_buffer_memory_limit; // protected by limit_lock
     uint64_t proxy_buffer_memory_used; // protected by limit_lock
@@ -836,10 +833,10 @@ struct conn {
     bool close_after_write; /** flush write then move to close connection */
     bool rbuf_malloced; /** read buffer was malloc'ed for ascii mget, needs free() */
     bool item_malloced; /** item for conn_nread state is a temporary malloc */
+    uint8_t ssl_enabled;
+    void    *ssl;
 #ifdef TLS
-    SSL    *ssl;
     char   *ssl_wbuf;
-    bool ssl_enabled;
 #endif
     enum conn_states  state;
     enum bin_substates substate;
@@ -948,6 +945,7 @@ extern void *ext_storage;
 /*
  * Functions
  */
+void verify_default(const char* param, bool condition);
 void do_accept_new_conns(const bool do_accept);
 enum delta_result_type do_add_delta(LIBEVENT_THREAD *t, const char *key,
                                     const size_t nkey, const bool incr,
